"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong

INDEX

Modules
-------
Index : Create and manage an array of dates
PrjtIndex : Create and manage index objects

Methods
-------
booleanloc : Return boolean array of data which is in array
"""

import pandas as pd
import numpy as np

#from pandas import (
#    date_range,
#    DatetimeIndex,
#    PeriodIndex,
    #RangeIndex,
#)

from datetime import (
    date,
    timedelta,
)
from dateutil.relativedelta import relativedelta

#from .genfunc import is_iterable
__all__ = ['RangeIndex', 'DateIndex', 'date_range', 'Index'] #, 'booleanloc', 'PrjtIndex']


def is_scalar(value):
    if isinstance(value, int):
        return True
    else:
        return False
        
def str_to_date(value):
    # date
    if isinstance(value, date):
        return value

    # date('-')
    value_lst = value.split('-')
    if len(value_lst) == 3:
        int_lst = [int(x) for x in value_lst]
        return date(*int_lst)
    elif len(value_lst) == 2:
        int_lst = [int(x) for x in value_lst]
        return date_monthend(*int_lst)
    
    # date('.')
    value_lst = value.split('.')
    if len(value_lst) == 3:
        int_lst = [int(x) for x in value_lst]
        return date(*int_lst)
    elif len(value_lst) == 2:
        int_lst = [int(x) for x in value_lst]
        return date_monthend(*int_lst)
        
    raise TypeError("Type of the data is not suitable.")
        
def date_monthend(year, month=None):
    if isinstance(year, date):
        dt = year
        return date_monthend(dt.year, dt.month)
        
    if isinstance(year, str):
        dt = year
        return date_monthend(str_to_date(dt))
    
    if month in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]:
        month_next = month + 1
        dt_next = date(year, month_next, 1)
        dt_monthend = dt_next - timedelta(days=1)
    elif month == 12:
        year_next = year + 1
        dt_next = date(year_next, 1, 1)
        dt_monthend = dt_next - timedelta(days=1)
    return dt_monthend

def date_next(dt, num=1, freq='M'):
    if isinstance(dt, str):
        dt = str_to_date(dt)
    assert isinstance(dt, date)
    if freq=='M':
        month_next = dt.month + num - 1
        year_next, month_next = divmod(month_next, 12)
        month_next += 1
        year_next = dt.year + year_next
        dt_next = date_monthend(year_next, month_next)
        return dt_next
    if freq=='Y':
        dt_next = dt + relativedelta(years=num)
        return dt_next
    if freq=='D':
        dt_next = dt + timedelta(days=num)
        return dt_next
    
def is_samemonth(*args):
    if len(args) < 2:
        raise ValueError("Number of values are not enough.")
    year_crit = args[0].year
    month_crit = args[0].month
    
    for val in args[1:len(args)]:
        if val.year != year_crit:
            return False
        if val.month != month_crit:
            return False
    return True

def all_none(*args):
    tmp_lst = [x==None for x in args]
    return all(tmp_lst)
    
def any_none(*args):
    tmp_lst = [x==None for x in args]
    return any(tmp_lst)

def allnot_none(*args):
    tmp_lst = [x!=None for x in args]
    return all(tmp_lst)
    
def anynot_none(*args):
    tmp_lst = [x!=None for x in args]
    return any(tmp_lst)


class RangeIndex():
    """
    Immutable sequence used for indexing a monotonic integer range.

    Parameters
    ----------
    start : int(default:0), range, RangeIndex instance
        If 'start' is an int and 'stop' is not given, interpreted as 'stop' instead.
    stop : int(default:0)
    step : int(default:1)
    name : str(default None)
        Name of the index.
    
    
    Attributes
    ----------
    name : The name of index
    len : The length of the index
    data : Raw data of index
    values : Raw data of index
    
    Methods
    -------
    copy : Deep copy of the index

    Examples
    --------
    Input one integer.
    >>> idxa = RangeIndex(10, name='IdxA')
    >>> idxa
        RangeIndex(range(0, 10))
    >>> idxa.name
        'IdxA'
    >>> len(idxa)
        10
    >>> idxa.len
        10
    >>> idxa.copy(name='NewIdxA')
        RangeIndex(range(0, 10))
    Input a RangeIndex.
    >>> RangeIndex(idxa)
        RangeIndex(range(0, 10))
    Input a range.
    >>> RangeIndex(range(0, 20))
        RangeIndex(range(0, 20))
    Input two integer(start, stop).
    >>> RangeIndex(10, 20)
        RangeIndex(range(10, 20))
    """
    def __new__(
        cls, 
        start=None,
        stop=None,
        step=None,
        name=None,
        ):
        
        # RangeIndex, range
        if isinstance(start, RangeIndex):
            if name is None:
                name = start.name
            return start.copy(name=name)
        elif isinstance(start, range):
            return cls._simple_new(start, name)
            
        # validate the arguments
        if all_none(start, stop, step):
            raise TypeError("RangeIndex(...) must be called with integers.")
            
        if start is not None:
            start = cls._ensure_int(start)
        else:
            start = 0
        
        if stop is None:
            (start, stop) = (0, start)
        else:
            stop = cls._ensure_int(stop)
            
        step = cls._ensure_int(step) if step is not None else 1
        if step == 0:
            raise ValueError("Step must not be zero.")
            
        rng = range(start, stop, step)
        
        return cls._simple_new(rng, name=name)
        
            
    @classmethod
    def _simple_new(cls, values: range, name=None):
        result = object.__new__(cls)
        
        assert isinstance(values, (range, RangeIndex))
        
        result._data = values
        result._range = values
        result._name = name
        result._cache = {}
        return result
        
    @classmethod
    def _ensure_int(cls, value):
        new_value = int(value)
        assert new_value == value
        return new_value
            
    def copy(self, name=None):
        try:
            casted = self._data
        except (TypeError, ValueError) as err:
            raise TypeError(
                f"Cannot cast {type(self).__name__}"
                ) from err
        if name is None:
            name = self.name
        return RangeIndex(casted, name=name)
    
    def __len__(self) -> int:
        """
        Return the length of the Index.
        """
        return len(self._data)
    
    def __setitem__(self, key, value):
        raise TypeError("Index does not support mutable operations")
    
    def __getitem__(self, key):
        getitem = self._data.__getitem__
        
        if is_scalar(key):
            return getitem(key)
        
        if isinstance(key, slice):
            result = getitem(key)
            return type(self)._simple_new(result, name=self.name)
        
    def __repr__(self):
        """
        Return a string representation for this object.
        """
        klass_name = type(self).__name__
        data = str(self._data)
        return f"{klass_name}({data})"
    
    @property
    def values(self):
        return self._data
    
    @property
    def data(self):
        return self._data
    
    @property
    def name(self):
        return self._name
    @name.setter
    def name(self, value):
        self._name = value

    @property
    def len(self):
        return self.__len__()
        

class DateIndex():
    """
    Immutable sequence used for indexing date.

    Parameters
    ----------
    data : date string, datetime.date, list(...)
    name : str(default None)
        Name of the index.

    Attributes
    ----------
    name : The name of index
    len : The length of the index
    data : Raw data of index
    values : Raw data of index

    Methods
    -------
    copy : Deep copy of the index

    Examples
    --------
    Input one date string.
    >>> DateIndex('2022.01')
        DateIndex(['2022.01.31'])
    >>> DateIndex("2022-3-27")
        DateIndex(['2022.03.27'])
    Input one date type data.
    >>> DateIndex(datetime.date(2022, 1, 1))
        DateIndex(['2022.01.01'])
    Input a list of date string.
    >>> DateIndex(['2022.01', '2022.02', '2022.03'])
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    Input a list of date type data.
    >>> DateIndex([date(2022, 1, 31), date(2022, 2, 28)])
        DateIndex(['2022.01.31', '2022.02.28'])
    """
    def __new__(
        cls,
        data=None,
        name=None,
        ):
        
        # DateIndex
        if isinstance(data, DateIndex):
            dtarr = data.copy(name=name)
            
        # string
        elif isinstance(data, str):
            dtarr = [str_to_date(data)]

        # date
        elif isinstance(data, date):
            dtarr = [data]
            
        # list
        elif isinstance(data, list):
            new_lst = []
            for val in data:
                if isinstance(val, date):
                    new_lst.append(val)
                elif isinstance(val, str):
                    new_lst.append(str_to_date(val))
            dtarr = new_lst
        else:
            raise TypeError("There is a problem with the entered data type.")
            
        subarr = cls._simple_new(dtarr, name=name)
        return subarr
        
    @classmethod
    def _simple_new(cls, values: list, name=None):
        result = object.__new__(cls)
        
        assert isinstance(values, (list, DateIndex))
        
        result._data = values
        result._name = name
        result._cache = {}
        return result
        
    def copy(self, name=None):
        try:
            casted = self._data
        except (TypeError, ValueError) as err:
            raise TypeError(
                f"Cannot cast {type(self).__name__}"
                ) from err
        if name is None:
            name = self.name
        return DateIndex(casted, name=name)        
    
    def __len__(self) -> int:
        """
        Return the length of the Index.
        """
        return len(self._data)
    
    def __setitem__(self, key, value):
        raise TypeError("Index does not support mutable operations")
    
    def __getitem__(self, key):
        getitem = self._data.__getitem__
        
        if is_scalar(key):
            return getitem(key)
        
        if isinstance(key, slice):
            result = getitem(key)
            return type(self)._simple_new(result, name=self.name)
        
    def __repr__(self):
        """
        Return a string representation for this object.
        """
        klass_name = type(self).__name__
        data = str([val.strftime("%Y.%m.%d") for val in self._data])
        return f"{klass_name}({data})"
    
    @property
    def values(self):
        return self._data
    
    @property
    def data(self):
        return self._data
    
    @property
    def name(self):
        return self._name
    @name.setter
    def name(self, value):
        self._name = value

    @property
    def len(self):
        return self.__len__()


def date_range(
    start=None,
    stop=None,
    periods=None,
    freq=None,
    name=None,
    ) -> DateIndex:
    """
    Return a fixed frequency DateIndex.
    
    Parameters
    ----------
    start : str or datetime.date
        Left bound for generating dates.
    stop : str or datetime.date
        Right bound for generating dates.
    periods : int(>0), optional
        Number of periods to generate.
    freq : str or DateOffset, default 'M'
    
    Returns
    -------
    DateIndex

    Attributes
    ----------
    name : The name of index
    len : The length of the index
    data : Raw data of index
    values : Raw data of index

    Methods
    -------
    copy : Deep copy of the index

    Examples
    --------
    Input a start date and periods.
    >>> date_range("2022.01", 3)
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    >>> date_range(date(2022, 1, 1), 3)
        DateIndex(['2022.01.01', '2022.02.28', '2022.03.31'])
    >>> idx = date_range("2022.01", 3)
    >>> idx.len
        3
    >>> idx.data
        [datetime.date(2022, 1, 1), datetime.date(2022, 2, 28), datetime.date(2022, 3, 31)]
    >>> idx.values
        [datetime.date(2022, 1, 1), datetime.date(2022, 2, 28), datetime.date(2022, 3, 31)]
    Input an stop date and periods.
    >>> date_range(stop='2022.12', periods=3)
        DateIndex(['2022.10.31', '2022.11.30', '2022.12.31'])
    Input a start date and an stop date.
    >>> date_range('2022.01', '2022.03')
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    """
    
    if freq is None and any_none(periods, start, stop):
        freq = "M"
    
    if allnot_none(start, stop, periods):
        raise ValueError("There are too many data.")
    elif allnot_none(start, periods):
        dtarr = []
        if periods > 0:
            start = str_to_date(start)
            dtarr.append(start)
            dt_next = start
        else:
            raise ValueError(
                "'periods' should be larger than 0."
            )
        for no in range(1, periods):
            dt_next = date_next(dt_next, freq=freq)
            dtarr.append(dt_next)
    elif allnot_none(stop, periods):
        stop = str_to_date(stop)
        start = stop - relativedelta(months=(periods-1))
        return date_range(start, periods=periods)
    elif allnot_none(start, stop):
        if isinstance(stop, int):
            periods = stop
            return date_range(start, periods=periods)
        else:
            dtarr = []
            start = str_to_date(start)
            stop = str_to_date(stop)
            if stop == start:
                dtarr.append(start)
            elif stop > start:
                dtarr.append(start)
                dt_next = start
                while True:
                    if dt_next >= stop:
                        break
                    dt_next = date_next(dt_next, freq=freq)
                    dtarr.append(dt_next)
            elif stop < start:
                raise ValueError("The stop date is before the start date.")
        
    subarr = DateIndex._simple_new(dtarr, name=name)
    return subarr


class Index():
    """
    Immutable sequence used for indexing.

    Parameters
    ----------
    start : int, str or datetime.date, list(date)
    stop : int, str or datetime.date
    step : int
    periods : int(>0), optional
        Number of periods to generate.
    freq : str or DateOffset, default 'M'
    name : name of index

    Returns
    -------
    RangeIndex or DateIndex

    Attributes
    ----------
    name : The name of index
    len : The length of the index
    data : Raw data of index
    values : Raw data of index

    Methods
    -------
    copy : Deep copy of the index

    Examples
    --------
    Input a start date and periods.
    >>> Index("2022.01", 3)
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    >>> Index(date(2022, 1, 1), 3)
        DateIndex(['2022.01.01', '2022.02.28', '2022.03.31'])
    >>> idx = Index("2022.01", 3)
    >>> idx.len
        3
    >>> idx.data
        [datetime.date(2022, 1, 1), datetime.date(2022, 2, 28), datetime.date(2022, 3, 31)]
    >>> idx.values
        [datetime.date(2022, 1, 1), datetime.date(2022, 2, 28), datetime.date(2022, 3, 31)]
    Input an stop date and periods.
    >>> Index(stop='2022.12', periods=3)
        DateIndex(['2022.10.31', '2022.11.30', '2022.12.31'])
    Input a start date and an stop date.
    >>> Index('2022.01', '2022.03')
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    Input one date string.
    >>> Index('2022.01')
        DateIndex(['2022.01.31'])
    >>> Index("2022-3-27")
        DateIndex(['2022.03.27'])
    Input one date type data.
    >>> Index(datetime.date(2022, 1, 1))
        DateIndex(['2022.01.01'])
    Input a list of date string.
    >>> Index(['2022.01', '2022.02', '2022.03'])
        DateIndex(['2022.01.31', '2022.02.28', '2022.03.31'])
    Input a list of date type data.
    >>> Index([date(2022, 1, 31), date(2022, 2, 28)])
        DateIndex(['2022.01.31', '2022.02.28'])
    Input one integer.
    >>> Index(10, name='IdxA')
        RangeIndex(range(0, 10))
    >>> idxa.name
        'IdxA'
    >>> idxa.copy(name='NewIdxA')
        RangeIndex(range(0, 10))
    Input a RangeIndex.
    >>> Index(idxa)
        RangeIndex(range(0, 10))
    Input a range.
    >>> Index(range(0, 20))
        RangeIndex(range(0, 20))
    Input two integer(start, stop).
    >>> Index(10, 20)
        RangeIndex(range(10, 20))
    """
    def __new__(
        cls,
        start=None,
        stop=None,
        step=None,
        periods=None,
        freq=None,
        name=None,
        ):

        # RangeIndex
        if isinstance(start, RangeIndex):
            if name is None:
                name = start.name
            return RangeIndex(start, name=name)
        if isinstance(start, range):
            return RangeIndex(start, name=name)
        if isinstance(start, int):
            return RangeIndex(start, stop, step, name)
        if start is None:
            if isinstance(stop, int):
                return RangeIndex(start, stop, step, name)

        # DateIndex
        if isinstance(start, DateIndex):
            if name is None:
                name = start.name
            return DateIndex(start, name=name)
        if isinstance(start, str):
            if all_none(stop, step, periods):
                return DateIndex(start, name=name)
        if isinstance(start, date):
            if all_none(stop, step, periods):
                return DateIndex(start, name=name)
        if isinstance(start, list):
            if all_none(stop, step, periods):
                return DateIndex(start, name=name)

        # date_range
        if isinstance(start, (str, date)):
            if isinstance(stop, (int, str, date)):
                return date_range(start, stop, periods=periods, freq=freq, name=name)
        if start is None:
            if isinstance(stop, (str, date)):
                return date_range(start, stop, periods=periods, freq=freq, name=name)




"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong
"""

import pandas as pd
from pandas import Series, DataFrame
pd.set_option('display.max_row', 200)
pd.set_option('display.max_columns', None)
pd.set_option('display.width', 1000)
import numpy as np
from functools import partial

from datetime import datetime
from datetime import date
from functools import wraps

from .genfunc import (
    limited,
    rounding,
    listwrapper,
)
from .index import (
    RangeIndex,
    DateIndex,
    date_range,
    str_to_date,
)

__all__ = ['Account', 'Setattr']

DFCOL      = ['scd_in', 'scd_in_cum', 'scd_out', 'scd_out_cum',
                           'bal_strt', 'amt_in', 'amt_in_cum',
                           'amt_out', 'amt_out_cum', 'bal_end',
                           'rsdl_in_cum', 'rsdl_out_cum']
DFCOL_smry = ['bal_strt', 'amt_in', 'amt_out', 'bal_end']
DFCOL_inpt = ['scd_in', 'scd_out', 'amt_in', 'amt_out']
JNLCOL     = ['amt_in', 'amt_out', 'rcvfrm', 'payto', 'note']

class Account:
    _index = RangeIndex(0)

    def __new__(cls, *args, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return cls._index_new(cls._index)

        if len(args) == 1 and len(kwargs) == 0:
            if isinstance(args[0], (DateIndex, RangeIndex)):
                return cls._index_new(args[0])
            if isinstance(args[0], DataFrame):
                return cls._df_new(args[0])


        if len(args) == 0 and len(kwargs) == 1:
            if 'index' in kwargs:
                if isinstance(kwargs['index'], (DateIndex, RangeIndex)):
                    return cls._index_new(kwargs['index'])
                else:
                    raise TypeError("Type of index should be an instance of DataIndex or RangeIndex.")
            if 'dataframe' in kwargs:
                if isinstance(kwargs['dataframe'], DataFrame):
                    return cls._df_new(kwargs['dataframe'])
                else:
                    raise TypeError("Type of dataframe should be an instance of DataFrame.")

    @classmethod
    def _index_new(cls, index):
        result = object.__new__(cls)
        result.index = index
        result.name = 'main'
        result._df = cls._make_dataframe(index)
        result._initialize()
        return result

    @classmethod
    def _df_new(cls, df):
        result = object.__new__(cls)
        result.index = df.index
        result.name = 'main'
        result._df = cls._make_dataframe(df.index)
        for key in df.columns:
            if key in DFCOL_inpt:
                result._df.loc[df.index, key] = df[key]
        result._initialize()
        return result

    @staticmethod
    def _make_dataframe(index, data=None):
        if data is None:
            data = np.zeros([len(index), len(DFCOL)])
        return DataFrame(data, columns=DFCOL, index=index)

    @staticmethod
    def _make_jnldf():
        return DataFrame(columns=JNLCOL)

    @staticmethod
    def _make_jnlscd():
        return DataFrame(columns=JNLCOL)

    def _initialize(self):
        self._jnl = self.__class__._make_jnldf()
        self._jnlscd = self.__class__._make_jnlscd()
        self._set_outputfunc()
        self._cal_bal()


    def __setattr__(self, key, value):
        if 'vars' not in self.__dict__:
            super().__setattr__('vars', [])
        super().__setattr__(key, value)

        if key in ['_df', '_jnl', '_jnlscd', '_dct']:
            return
        if key in DFCOL:
            return
        if key not in self.vars:
            self.vars.append(key)

    def __repr__(self):
        """Return a string representation for this object."""
        if '_dct' in vars(self):
            repr = f"Account({self.name}, len {len(self.index)!r}, dct: {list(self.dct.keys())!r})"
        else:
            repr = f"Account({self.name}, len {len(self.index)!r})"

        """if len(self.df) > 5:
            repr = f"{self.df.head(2)}\n...\n{self.df.tail(2)}"
        else:
            repr = f"{self.df}"
        if 'name' in self.__dict__:
            repr = f"<{self.name}>\n" + repr"""
        return repr

    def __enter__(self):
        return self
    def __exit__(self, type, value, trackback):
        pass

    #Calculate Data Balance
    def _cal_bal(self):
        if len(self.index) == 0:
            return None

        self._df.scd_in_cum = self._df.scd_in.cumsum()
        self._df.scd_out_cum = self._df.scd_out.cumsum()
        self._df.amt_in_cum = self._df.amt_in.cumsum()
        self._df.amt_out_cum = self._df.amt_out.cumsum()

        #Calculate account balance
        for i, idx in enumerate(self.index):
            if i > 0:
                self._df.loc[idx, 'bal_strt'] = self._df.loc[idxpst, 'bal_end']
            self._df.loc[idx, 'bal_end'] = (self._df.loc[idx, 'bal_strt']
                                           + self._df.loc[idx, 'amt_in']
                                           - self._df.loc[idx, 'amt_out'])
            idxpst = idx

        # Calculate the residual of scheduled amounts
        self._df.loc[:, 'rsdl_in_cum'] = self._df.loc[:, 'scd_in_cum'] \
                                         - self._df.loc[:, 'amt_in_cum']
        self._df.loc[:, 'rsdl_out_cum'] = self._df.loc[:, 'scd_out_cum'] \
                                          - self._df.loc[:, 'amt_out_cum']

    #Input Data
    @listwrapper
    def addscd(self, idxval, amt, rcvfrm=None, note='add_scd'):
        """
        Add the amount on the 'scd_in' columns of dataframe.

        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        rcvfrm : str
        note : str

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        self.iptjnlscd(idxval, amt, 0, rcvfrm, None, note)
        self._df.loc[idxval, 'scd_in'] += amt
        self._cal_bal()

    @listwrapper
    def subscd(self, idxval, amt, payto=None, note='sub_scd'):
        """
        Add the amount on the 'scd_out' columns of dataframe.

        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        payto : str
        note : str

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        self.iptjnlscd(idxval, 0, amt, None, payto, note)
        self._df.loc[idxval, 'scd_out'] += amt
        self._cal_bal()

    @listwrapper
    def addamt(self, idxval, amt, rcvfrm=None, note='add_amt'):
        """
        Add the amount on the 'amt_in' columns of dataframe.

        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        rcvfrm : str
        note : str

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        if amt == 0:
            return
        self.iptjnl(idxval, amt, 0, rcvfrm, None, note)
        self._df.loc[idxval, 'amt_in'] += amt
        self._cal_bal()

    @listwrapper
    def subamt(self, idxval, amt, payto=None, note='sub_amt'):
        """
        Add the amount on the 'amt_out' columns of dataframe.

        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        payto : str
        note : str

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        if amt == 0:
            return
        self.iptjnl(idxval, 0, amt, None, payto, note)
        self._df.loc[idxval, 'amt_out'] += amt
        self._cal_bal()

    @listwrapper
    def iptamt(self, idxval, amt, rcvfrm=None, payto=None, note=None):
        """
        If the amount is positive, apply the addamt, else apply the subamt

        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        rcvfrm : str, default None
        payto : str, default None
        note : str, default "sub_amt"

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        if amt == 0:
            return
        if amt > 0:
            if rcvfrm is None:
                rcvfrm = "add_amt"
            self.addamt(idxval, amt, rcvfrm, note)
        else:
            if payto is None:
                payto = "sub_amt"
            self.subamt(idxval, -amt, payto, note)

    def iptjnlscd(self, idxval, amt_in, amt_out, rcvfrm=None, payto=None, note=None):
        """
        Add the amount on the journal schedule.

        Parameters
        ----------
        idxval
        amt_in
        amt_out
        rcvfrm
        payto
        note

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        tmpjnl = DataFrame([[amt_in, amt_out, rcvfrm, payto, note]],
                           columns=JNLCOL,
                           index=[idxval])
        self._jnlscd = pd.concat([self._jnlscd, tmpjnl])

    def iptjnl(self, idxval, amt_in, amt_out, rcvfrm=None, payto=None, note=None):
        """
        Add the amount on the journal.

        Parameters
        ----------
        idxval : index
        amt_in : int, float
        amt_out : int, float
        rcvfrm : str, default None
        payto : str, default None
        note : str, default None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        tmpjnl = DataFrame([[amt_in, amt_out, rcvfrm, payto, note]],
                           columns=JNLCOL, index=[idxval])
        self._jnl = pd.concat([self._jnl, tmpjnl])

    #Output Data
    @property
    def df(self):
        """
        Return the summarised dataframe

        Returns
        -------
        Return columns: DFCOL_smry
            ['bal_strt', 'amt_in', 'amt_out', 'bal_end']
        """
        return self._df.loc[:, DFCOL_smry]

    @property
    def dfall(self):
        """
        Return the all dataframe
        """
        return self._df

    @property
    def jnl(self):
        """
        Return the journal dataframe
        """
        return self._jnl
    @jnl.setter
    def jnl(self, jnlnew):
        if not isinstance(jnlnew, DataFrame):
            raise AttributeError
        if set(jnlnew.columns) != set(JNLCOL):
            raise AttributeError
        self._jnl = jnlnew

    @property
    def jnlscd(self):
        """
        Return the journal schedule dataframe
        """
        return self._jnlscd
    @jnlscd.setter
    def jnlscd(self, jnlnew):
        if not isinstance(jnlnew, DataFrame):
            raise AttributeError
        if set(jnlnew.columns) != set(JNLCOL):
            raise AttributeError
        self._jnlscd = jnlnew

    #Decorator
    class getattr_dfcol:
        """
        Decorator
        Get a class name and use the class name as the column of dataframe.
        """

        def __call__(self, cls):
            def init(self, spristnc):
                self.spristnc = spristnc
                self.colname = cls.__name__

            cls.__init__ = init

            def getitem(self, val):
                """
                If val is an integer, return the data which is in index[val].
                If val is a date, return the data which is on the date.
                If val is a string, get the date data and return the data
                    which is on the date.

                Parameters
                ----------
                val: int, slice, date, date like string,
                    ex) 0, 1:3, datetime.date(2021, 4, 30), "2021-04", "2021-04-30"

                Return
                ------
                data from dataframe
                Array of data from dataframe

                Examples
                --------
                >>> idx = DateIndex("2021.01", "2021.12")
                >>> acc = Account(idx, "loan")
                >>> acc.addscd(idx[0], 1000)
                >>> acc.subscd(idx[5], 800)
                >>> acc.addamt(idx[1], 500, "acc_oprtg", "amount acc oprtg")
                >>> acc.scd_in[idx[3]]
                    0.0
                >>> acc.scd_in[3]
                    0.0
                >>> acc.scd_in[0:3]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in[datetime.date(2021, 3, 31)]
                    0.0
                >>> acc.scd_in["2021.04"]
                    2021-04-30    0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in["2021"]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    2021-04-30       0.0
                    2021-05-31       0.0
                    2021-06-30       0.0
                    2021-07-31       0.0
                    2021-08-31       0.0
                    2021-09-30       0.0
                    2021-10-31       0.0
                    2021-11-30       0.0
                    Name: scd_in, dtype: float64
                """
                if isinstance(val, date):
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, int):
                    val = self.spristnc.index[val]
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, str):
                    val = str_to_date(val)
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, slice):
                    if isinstance(val.start, str):
                        new_start = str_to_date(val.start)
                    elif isinstance(val.start, int):
                        new_start = self.spristnc.index[val.start]
                    elif isinstance(val.start, date):
                        new_start = val.start

                    if isinstance(val.stop, str):
                        new_stop = str_to_date(val.stop)
                    elif isinstance(val.stop, int):
                        new_stop = self.spristnc.index[val.stop]
                    elif isinstance(val.stop, date):
                        new_stop = val.stop

                    new_slice = slice(new_start, new_stop)
                    return self.spristnc._df.loc[new_slice, self.colname]
            cls.__getitem__ = getitem

            return cls

    @getattr_dfcol()
    class scd_in:
        pass

    @getattr_dfcol()
    class scd_in_cum:
        pass

    @getattr_dfcol()
    class scd_out:
        pass

    @getattr_dfcol()
    class scd_out_cum:
        pass

    @getattr_dfcol()
    class bal_strt:
        pass

    @getattr_dfcol()
    class amt_in:
        pass

    @getattr_dfcol()
    class amt_in_cum:
        pass

    @getattr_dfcol()
    class amt_out:
        pass

    @getattr_dfcol()
    class amt_out_cum:
        pass

    @getattr_dfcol()
    class bal_end:
        pass

    @getattr_dfcol()
    class rsdl_in_cum:
        pass

    @getattr_dfcol()
    class rsdl_out_cum:
        pass

    def _set_outputfunc(self):
        self.scd_in = self.scd_in(self)
        self.scd_in_cum = self.scd_in_cum(self)
        self.scd_out = self.scd_out(self)
        self.scd_out_cum = self.scd_out_cum(self)
        self.bal_strt = self.bal_strt(self)
        self.amt_in = self.amt_in(self)
        self.amt_in_cum = self.amt_in_cum(self)
        self.amt_out = self.amt_out(self)
        self.amt_out_cum = self.amt_out_cum(self)
        self.bal_end = self.bal_end(self)
        self.rsdl_in_cum = self.rsdl_in_cum(self)
        self.rsdl_out_cum = self.rsdl_out_cum(self)

    def amt_rqrd_excs(self, idxval, rqrdamt, minunit=100):
        """
        Calculate the additional amount required in excess of the balance.

        Parameters
        ----------
        index : index number
        rqrdamt : int, float, total required amount
        minunit : int, default 100, minimum adjustment unit

        Returns
        -------
        amt_rqrd : float
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        amt_rqrd = max(rqrdamt - self.bal_end[idxval], 0)
        amt_rqrd = round_up(amt_rqrd, -log10(minunit))
        return amt_rqrd

    def send(self, idxval, amt, account, note=None):
        """
        Transfer the amount from this account to the opponent account.

        Parameters
        ----------
        index : index
        amt : int, float, amount to transfer
        account : account
        note : str, default None

        Returns
        -------
        None
        """
        if isinstance(idxval, int):
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)

        if 'name' in account.__dict__:
            account_name = account.name
        else:
            account_name = None
        self.subamt(idxval, amt, account_name, note)

        if 'name' in self.__dict__:
            self_name = self.name
        else:
            self_name = None
        account.addamt(idxval, amt, self_name, note)

    #Set and Manage Sub Account
    def subacc(self, name):
        _acc = Account(self.index)
        _acc.name = name

        if '_dct' not in vars(self):
            self._dct = {}

        self._dct[name] = _acc
        return _acc

    @property
    def dct(self):
        return vars(self).get('_dct', None)

    @property
    def mrg(self):
        dflst = [item._df for item in self._dct.values()]
        newacc = Account(sum(dflst))

        jnllst = [item.jnl for item in self._dct.values()]
        jnlmrg = pd.concat(jnllst).sort_index()

        jnlscdlst = [item.jnlscd for item in self._dct.values()]
        jnlscdmrg = pd.concat(jnlscdlst).sort_index()

        newacc._jnl = jnlmrg
        newacc._jnlscd = jnlscdmrg

        return newacc

    @property
    def mrg2(self):
        accdct = {key: item.mrg for key, item in self._dct.items()}

        dflst = [item._df for item in accdct.values()]
        newacc = Account(sum(dflst))

        jnllst = [item.jnl for item in accdct.values()]
        jnlmrg = pd.concat(jnllst).sort_index()

        jnlscdlst = [item.jnlscd for item in accdct.values()]
        jnlscdmrg = pd.concat(jnlscdlst).sort_index()

        newacc._jnl = jnlmrg
        newacc._jnlscd = jnlscdmrg

        return newacc


#Decorator
class Setattr:
    def __init__(self, ipt):
        self.ipt = ipt
    def __call__(self, func):
        if isinstance(self.ipt, Account):
            setattr(self.ipt, func.__name__, partial(func, self.ipt))
        elif isinstance(self.ipt, dict):
            for item in self.ipt.values():
                setattr(item, func.__name__, partial(func, item))


"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong

General functions for cafle.

Attributes
----------
PY : the area unit used in Korea

Modules
-------
Area : transform between m2 and PY units. 
"""

import math
from datetime import datetime
from functools import wraps
from collections import OrderedDict
from pandas.api.types import is_numeric_dtype

__all__ = ['PY', 'Area', 'EmptyClass', 'is_iterable', 'limited', 'rounding', 
           'print_rounding', 'round_up', 'log10', 'extnddct']

PY = 1/3.305785
# PY("pyung") is the area unit used in Korea.

class Area:
    """
    Transform between m2 and PY units.
    
    Parameters
    ----------
    m2 : int, float, default None
    py : int, float, default None
    roundunit : int, default 2
    
    Attributes
    ----------
    area : tuple
        return a tuple of m2, py.
    m2 : float
        value of m2 unit
    py : float
        value of py unit
        
    Examples
    --------
    >>> ar = Area(1000)
    >>> ar.area
        (1000, 302.5)
    >>> ar.py
        302.5
    >>> ar.m2
        1000
    """
    def __init__(self, m2=None, py=None, roundunit=2):
        self._m2 = m2
        self._py = py
        self._roundunit = roundunit
        self._intlz()
        
    def _intlz(self):
        if all([self._m2 is not None, self._py is None]):
            self._py = round(self._m2 * PY, self._roundunit)
        if all([self._m2 is None, self._py is not None]):
            self._m2 = round(self._py / PY, self._roundunit)
            
    def __repr__(self):
        repr_tmp = f"{self.m2:,.0f}m2 ({self.py:,.0f}py)"
        return repr_tmp
    
    @property
    def area(self):
        return (self._m2, self._py)
        
    @property
    def m2(self):
        return self._m2
        
    @property
    def py(self):
        return self._py
        
        
class EmptyClass:
    def __init__(self):
        pass
        
    def __getattr__(self, attr):
        return self.__dict__[attr]
        
    def __getitem__(self, key):
        return self.__dict__[key]

    @property
    def dict(self):
        return self.__dict__

    @property
    def keys(self):
        return self.__dict__.keys()
        

def is_scalar(data):
    if isinstance(data, (str, int, float)):
        return True
    else:
        return False


def is_iterable(data):
    if type(data) == str:
        return False
    try:
        _ = iter(data)
        return True
    except TypeError:
        return False
        
        
def limited(val, upper=None, lower=None):
    """
    Adjust the input value between "upper" and "lower" values.
    
    Parameters
    ----------
    val : int, float
        input value
    upper : int, float, list, tuple
        the criteria for distinguishing upper values
    lower : int, float, list, tuple
        the criteria for distinguishing lower values
    
    Return
    ------
    Adjusted value between upper values and lower values.
        
    Examples
    --------
    >>> limited(100, upper=90, lower=50)
        90
    >>> limited(30, lower=[10, 40])
        40
    """

    tmp_val = val
    
    if upper is not None:
        if is_iterable(upper):
            for val_lmt in upper:
                tmp_val = min(tmp_val, val_lmt)
        else:
            tmp_val = min(tmp_val, upper)
            
    if lower is not None:
        if is_iterable(lower):
            for val_lmt in lower:
                tmp_val = max(tmp_val, val_lmt)
        else:
            tmp_val = max(tmp_val, lower)
    return tmp_val
    
    
def rounding(df, rate=None):
    rslt_df = df.copy()
    for key, item in rslt_df.items():
        if rate is not None and key in rate:
            rslt_df[key] = rslt_df[key] * 100
        if all([isinstance(val, (int, float)) for val in rslt_df[key]]):
            rslt_df[key] = rslt_df[key].fillna(0).apply(lambda x: f"{x:,.0f}")
    return rslt_df
    
    
def print_rounding(df):
    """
    Apply a separator to the number and round the number.
    
    Parameters
    ----------
    df : DataFrame
    
    Return
    ------
    DataFrame rounded and applied a separator to the number
        
    Examples
    --------
    >>> df = DataFrame({'a':[100000,     200000    ], 
                        'b':[100000.123, 200000.321], 
                        'c':['abc',      'cde'     ]})
    >>> print_rounding(df)
                 a        b    c
        0  100,000  100,000  abc
        1  200,000  200,000  cde
    """
    rslt_df = df.copy()
    for key, item in rslt_df.items():
        if is_numeric_dtype(item):
            rslt_df[key] = rslt_df[key].fillna(0).apply(lambda x: f"{x:,.0f}")
    return rslt_df


def margindf(iptval):
    """
    This method returns DataFrame or Series with sum value on the margin of it.

    Parameters
    ----------
    iptval : DataFrame, Series

    Returns
    -------
    DataFrame, Series
    """
    if type(iptval) == DataFrame:
        rsltdf = iptval.copy()
        sumdf = DataFrame(columns=rsltdf.columns, index=['Sum'])

        for key, item in rsltdf.items():
            if is_numeric_dtype(item):
                dtype = item.dtype
                sumval = item.sum()
                sumdf.loc['Sum', key] = sumval
                sumdf[key] = sumdf[key].astype(dtype)
            else:
                sumdf.loc['Sum', key] = '-'
        rsltdf = pd.concat([rsltdf, sumdf])
        return rsltdf

    if type(iptval) == Series:
        rsltsrs = iptval.copy()
        if is_numeric_dtype(rsltsrs):
            dtype = rsltsrs.dtype
            sumval = rsltsrs.sum()
        else:
            sumval = '-'
        sumsrs = Series(sumval, index=['Sum']).astype(dtype)
        rsltsrs = pd.concat([rsltsrs, sumsrs])
        return rsltsrs


def showdf(iptval, margins=False):
    """
    This method shows DataFrame or Series more easily understood.

    Parameters
    ----------
    iptval : DataFrame or Series
    margins : True or False

    Returns
    -------
    DataFrame or Series rounded and applied a separator to the number.
    And margins option makes sum row on the margin.

    Examples
    --------
    >>> df = DataFrame({'a':[100000,     200000    ],
                        'b':[100000.123, 200000.321],
                        'c':['abc',      'cde'     ]})
    >>> print_rounding(df)
                 a        b    c
        0  100,000  100,000  abc
        1  200,000  200,000  cde
    """
    if margins is True:
        iptval = margindf(iptval).copy()

    if type(iptval) == DataFrame:
        rsltdf = iptval.copy()
        for key, item in rsltdf.items():
            if is_numeric_dtype(item):
                rsltdf[key] = rsltdf[key].fillna(0).apply(lambda x: f"{x:,.0f}")
        return rsltdf

    if type(iptval) == Series:
        rsltsrs = iptval.copy()
        if is_numeric_dtype(rsltsrs):
            rsltsrs = rsltsrs.fillna(0).apply(lambda x: f"{x:,.0f}")
        return rsltsrs


def round_up(number:float, decimals:int=2):
    """
    Return a value rounded up to a specific number of decimal places.
    
    Parameters
    ----------
    number : int, float
    decimals : int
    
    Return
    ------
    float
        
    Examples
    --------
    >>> round_up(123.1231, 1)
        123.2
    """
    if not isinstance(decimals, int):
        raise TypeError("decimal places must be an integer")
    elif decimals == 0:
        return math.ceil(number)
        
    factor = 10 ** decimals
    return math.ceil(number * factor) / factor

def log10(val):
    tmpval = 0
    while True:
        val = val / 10
        if val > 0.9:
            tmpval += 1
        else:
            return tmpval

# Decorator
def listwrapper(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        ilen = 1
        for arg in args:
            if is_iterable(arg) is True:
                if ilen == 1:
                    ilen = len(arg)
                else:
                    if ilen != len(arg):
                        raise ValueError("All arguments should be of the same length.")
        for item in kwargs.values():
            if is_iterable(item) is True:
                if ilen == 1:
                    ilen = len(item)
                else:
                    if ilen != len(item):
                        raise ValueError("All arguments should be of the same length.")
        for i in range(ilen):
            new_args = []
            new_kwargs = {}
            for val in args:
                if is_iterable(val) is True:
                    new_args.append(val[i])
                else:
                    new_args.append(val)
            new_args = tuple(new_args)
            for key, item in kwargs.items():
                if is_iterable(item) is True:
                    new_kwargs[key] = item[i]
                else:
                    new_kwargs[key] = item
            func(self, *new_args, **new_kwargs)
    return wrapped

# Extend Dictionary
def extnddct(dct, *args):
    #dct = OrderedDict(dct)
    for val in args:
        #val = OrderedDict(val)
        dct.update(val)
    return dct




"""
Created on Sun Dec 19 17:41:35 2021
Editted on Jul 24 2022
Editted on Oct 12 2022

@author: KP_Hong
"""

import xlsxwriter
from datetime import datetime
from pandas import DataFrame
from cafle.genfunc import is_iterable, is_scalar
from collections import namedtuple

__all__ = ['Write', 'WriteWS', 'Cell']

ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
alphabet = "abcdefghijklmnopqrstuvwxyz"

class Cell:
    def __new__(cls, *args):
        if len(args) == 2:
            if isinstance(args[0], int) and isinstance(args[1], int):
                return cls._cell_rowncol(*args)

    @classmethod
    def _cell_rowncol(cls, row: int, col: int):
        result = object.__new__(cls)
        result.row = row
        result.col = col
        return result
    def __repr__(self):
        repr = f"Cell({self.row}, {self.col})"
        return repr

    @property
    def strfmt(self):
        numcol = self.col
        strcol = ""

        while True:
            quotient = numcol // 26
            remainder = numcol % 26
            strcol = ALPHABET[remainder] + strcol
            if quotient == 0:
                break
            numcol = quotient - 1

        strrow = self.row + 1
        return f"{strcol}{strrow}"


class Write:
    def __init__(self, file_adrs):
        self.file_adrs = file_adrs
        self.wb = xlsxwriter.Workbook(file_adrs)
        self.ws = {}

    def close(self):
        self.wb.close()
        return True

    def write_val(self, cell, val, ws, fmt=None, nxtcell='col'):
        ws = self.return_ws(ws)
        ws.write(cell.row, cell.col, val, fmt)
        row = cell.row
        col = cell.col

        if nxtcell == 'col':
            row += 1
        elif nxtcell == 'row':
            col += 1
        return Cell(row, col)

    def write_row(self, cell, data, ws, fmt=None, nxtcell='col'):
        ws = self.return_ws(ws)
        row = cell.row
        col = cell.col
        dlen = 0

        if not is_iterable(fmt):
            fmt = [fmt] * len(data)

        if not is_iterable(data[0]):
            _col = col
            for val, _fmt in zip(data, fmt):
                ws.write(row, _col, val, _fmt)
                _col += 1
            dlen = max(len(data), dlen)
            row += 1
        elif is_iterable(data[0]):
            for val in data:
                self.write_row(Cell(row, col), val, ws, fmt, nxtcell)
                dlen = max(len(val), dlen)
                row += 1

        if nxtcell == 'row':
            row = cell.row
            col = cell.col + dlen
        return Cell(row, col)

    def write_col(self, cell, data, ws, fmt=None, nxtcell='row'):
        ws = self.return_ws(ws)
        row = cell.row
        col = cell.col
        dlen = 0

        if not is_iterable(fmt):
            fmt = [fmt] * len(data)

        if not is_iterable(data[0]):
            _row = row
            for val, _fmt in zip(data, fmt):
                ws.write(_row, col, val, _fmt)
                _row += 1
            dlen = max(len(data), dlen)
            col += 1
        elif is_iterable(data[0]):
            for val in data:
                self.write_col(Cell(row, col), val, ws, fmt, nxtcell)
                dlen = max(len(val), dlen)
                col += 1

        if nxtcell == 'col':
            row = cell.row + dlen
            col = cell.col
        return Cell(row, col)

    def write_dct_col(self, cell, data, ws, fmt=None, fmtkey=None):
        ws = self.return_ws(ws)
        row = cell.row
        col = cell.col

        for key, item in data.items():
            if isinstance(fmt, dict):
                _fmt = fmt[key]
            else:
                _fmt = fmt
            if not isinstance(item, dict):
                self.write_val(Cell(row, col), key, ws, fmtkey)
                if is_iterable(item):
                    self.write_row(Cell(row, col + 1), item, ws, _fmt)
                    row += 1
                else:
                    self.write_val(Cell(row, col + 1), item, ws, _fmt)
                    row += 1
            elif isinstance(item, dict):
                self.write_val(Cell(row, col), key, ws, fmtkey)
                _cell = self.write_dct_col(Cell(row, col + 1), item, ws, _fmt, fmtkey)
                row = _cell.row
        return Cell(row, col)

    def write_dct_row(self, cell, data, ws, fmt=None, fmtkey=None):
        ws = self.return_ws(ws)
        row = cell.row
        col = cell.col

        for key, item in data.items():
            if isinstance(fmt, dict):
                _fmt = fmt[key]
            else:
                _fmt = fmt
            if not isinstance(item, dict):
                self.write_val(Cell(row, col), key, ws, fmtkey)
                if is_iterable(item):
                    self.write_col(Cell(row + 1, col), item, ws, _fmt)
                    col += 1
                else:
                    self.write_val(Cell(row + 1, col), item, ws, _fmt)
                    col += 1
            elif isinstance(item, dict):
                self.write_val(Cell(row, col), key, ws, fmtkey)
                _cell = self.write_dct_row(Cell(row + 1, col), item, ws, _fmt, fmtkey)
                col = _cell.col
        return Cell(row, col)

    def write_df_row(self, cell, data, ws, fmt=None, fmtkey=None, fmtidx=None):
        ws = self.return_ws(ws)
        dfdict = data.to_dict()
        idx = data.index
        row = cell.row
        col = cell.col

        self.write_col(Cell(row + 1, col), idx, ws, fmtidx)
        col += 1

        for key, item in dfdict.items():
            if isinstance(fmt, dict):
                _fmt = fmt[key]
            else:
                _fmt = fmt

            self.write_val(Cell(row, col), key, ws, fmtkey)
            for i, val in enumerate(idx, 1):
                self.write_val(Cell(row + i, col), item[val], ws, _fmt)
            col += 1
        return Cell(row, col)

    def write_df_col(self, cell, data, ws, fmt=None, fmtkey=None, fmtidx=None):
        ws = self.return_ws(ws)
        dfdict = data.to_dict()
        idx = data.index
        row = cell.row
        col = cell.col

        self.write_row(Cell(row, col + 1), idx, ws, fmtidx)
        row += 1

        for key, item in dfdict.items():
            if isinstance(fmt, dict):
                _fmt = fmt[key]
            else:
                _fmt = fmt

            self.write_val(Cell(row, col), key, ws, fmtkey)
            for i, val in enumerate(idx, 1):
                self.write_val(Cell(row, col + i), item[val], ws, _fmt)
            row += 1
        return Cell(row, col)

    #Make Dictionary of Loan
    def dct_loan(self, ln):
        tmpdct = {} #OrderedDict()
        tmpdct["Notional_" + ln.name] = \
            {"scd_out": ln.ntnl._df.scd_out,
             "scd_in": ln.ntnl._df.scd_in,
             "amt_out": ln.ntnl._df.amt_out,
             "amt_in": ln.ntnl._df.amt_in,
             "bal_end": ln.ntnl._df.bal_end}

        if 'IR' in ln.dct.keys():
            tmpdct["IR_" + ln.name] = \
                {"amt_in": ln.IR._df.amt_in,
                 "bal_end": ln.IR._df.bal_end}
        if 'fee' in ln.dct.keys():
            tmpdct["Fee_" + ln.name] = \
                {"amt_in": ln.fee._df.amt_in,
                 "bal_end": ln.fee._df.bal_end}
        if 'fob' in ln.dct.keys():
            tmpdct["Fob_" + ln.name] = \
                {"amt_in": ln.fob._df.amt_in,
                 "bal_end": ln.fob._df.bal_end}
        return tmpdct

    # Make Dictionary of Loan
    def dctprt_loan(self, ln):
        tmpdct = {} #OrderedDict()
        tmpdct["Notional_" + ln.name] = \
            {"인출한도": ln.ntnl._df.scd_out,
             "상환예정": ln.ntnl._df.scd_in,
             "인출금액": ln.ntnl._df.amt_out,
             "상환금액": ln.ntnl._df.amt_in,
             "대출잔액": ln.ntnl._df.bal_end}

        if 'IR' in ln.dct.keys():
            tmpdct["IR_" + ln.name] = \
                {"이자금액": ln.IR._df.amt_in,
                 "누적이자": ln.IR._df.bal_end}
        if 'fee' in ln.dct.keys():
            tmpdct["Fee_" + ln.name] = \
                {"수수료금액": ln.fee._df.amt_in,
                 "누적수수료": ln.fee._df.bal_end}
        if 'fob' in ln.dct.keys():
            tmpdct["미인출_" + ln.name] = \
                {"미인출수수료": ln.fob._df.amt_in,
                 "누적미인출": ln.fob._df.bal_end}
        return tmpdct

    def add_ws(self, wsname):
        self.ws[wsname] = self.wb.add_worksheet(wsname)
        writews = WriteWS(self.ws[wsname], cell=Cell(0,0))
        return writews

    def return_ws(self, wsname):
        if isinstance(wsname, str):
            return self.ws[wsname]
        elif isinstance(wsname, xlsxwriter.worksheet.Worksheet):
            return wsname

    #Text Format
    def fmtnum(self, fmt, **kwargs):
        """
        example of fmt: 'yyyy-mm-dd', '#,##0', '#,##0.0', '0.0%'
        kwargs example: bold=True
        """
        tmpdct = dict({'num_format': fmt, **kwargs})
        return self.wb.add_format(tmpdct)

    @property
    def nml(self):
        return self.wb.add_format({'bold': False})

    @property
    def bold(self):
        return self.wb.add_format({'bold': True})

    @property
    def num(self):
        return self.wb.add_format({'num_format': '#,##0'})

    @property
    def numb(self):
        return self.wb.add_format({'num_format': '#,##0', 'bold': True})

    @property
    def pct(self):
        return self.wb.add_format({'num_format': '0.0%'})

    @property
    def pct2(self):
        return self.wb.add_format({'num_format': '0.00%'})

    @property
    def date(self):
        return self.wb.add_format({'num_format': 'yyyy-mm-dd'})

    @property
    def month(self):
        return self.wb.add_format({'num_format': '#,##0"개월"'})

    @property
    def now(self):
        return datetime.now().strftime('%Y.%m.%d %H:%M:%S')


class WriteWS(Write):
    def __init__(self, ws, cell=Cell(0,0)):
        self.ws = ws
        self.cell = cell

    def __call__(self, data, fmt=None, fmtkey=None, fmtidx=None, valdrtn='row', drtn='col'):
        if not is_iterable(data):
            self.cell = self.write_val(self.cell, data, self.ws, fmt=fmt, nxtcell=drtn)
        elif isinstance(data, DataFrame):
            if valdrtn == 'row':
                if drtn == 'row':
                    self.cell = self.write_df_row(self.cell, data, self.ws, fmt, fmtkey, fmtidx)
                elif drtn == 'col':
                    _cell = self.write_df_row(self.cell, data, self.ws, fmt, fmtkey, fmtidx)
                    _len = len(data)
                    self.cell = Cell(self.cell.row + _len + 1, self.cell.col)
            elif valdrtn == 'col':
                if drtn == 'col':
                    self.cell = self.write_df_col(self.cell, data, self.ws, fmt, fmtkey, fmtidx)
                elif drtn == 'row':
                    _cell = self.write_df_col(self.cell, data, self.ws, fmt, fmtkey, fmtidx)
                    _len = len(data)
                    self.cell = Cell(self.cell.row, self.cell.col + _len + 1)
        elif is_iterable(data) and not isinstance(data, dict):
            if valdrtn == 'row':
                self.cell = self.write_row(self.cell, data, self.ws, fmt=fmt, nxtcell=drtn)
            if valdrtn == 'col':
                self.cell = self.write_col(self.cell, data, self.ws, fmt=fmt, nxtcell=drtn)
        elif isinstance(data, dict):
            if valdrtn == 'row':
                if drtn == 'row':
                    self.cell = self.write_dct_row(self.cell, data, self.ws, fmt=fmt, fmtkey=fmtkey)
                elif drtn == 'col':
                    _cell = self.write_dct_row(self.cell, data, self.ws, fmt=fmt, fmtkey=fmtkey)
                    _len = self.dctdatalen(data)
                    self.cell = Cell(self.cell.row + _len + 1, self.cell.col)
            if valdrtn == 'col':
                if drtn == 'col':
                    self.cell = self.write_dct_col(self.cell, data, self.ws, fmt=fmt, fmtkey=fmtkey)
                elif drtn == 'row':
                    _cell = self.write_dct_col(self.cell, data, self.ws, fmt=fmt, fmtkey=fmtkey)
                    _len = self.dctdatalen(data)
                    self.cell = Cell(self.cell.row, self.cell.col + _len + 1)
        return self.cell

    def setcell(self, row=None, col=None):
        if isinstance(row, Cell):
            self.cell = row
            return None
        if row is not None:
            _row = row
        if col is not None:
            _col = col
        self.cell = Cell(_row, _col)
        return None

    def nextcell(self, no=1, drtn='row'):
        _row = self.cell.row
        _col = self.cell.col
        if drtn == 'row':
            _row += no
        elif drtn == 'col':
            _col += no
        self.cell = Cell(_row, _col)
        return self.cell

    def dctdatalen(self, dct):
        _len = 0
        for key, item in dct.items():
            if not isinstance(item, dict):
                if is_iterable(item):
                    _len = max(_len, len(item))
                else:
                    _len = max(_len, 1)
            elif isinstance(item, dict):
                _len = max(_len, 1 + self.dctdatalen(item))
        return _len


